import numpy as np
from collections import OrderedDict
from string import Template

from ..resources import get_resource

def load_ply(file_obj, *args, **kwargs):
    '''
    Load a PLY file from an open file object.
    
    Arguments
    ---------
    file_obj: an open file- like object

    Returns
    ---------
    mesh_kwargs: dictionary of mesh info which can be passed to 
                 Trimesh constructor, eg: a = Trimesh(**mesh_kwargs)
    '''
    def ply_element_colors(element):
        '''
        Given an element, try to extract RGBA color from its properties
        and return them as an (n,3|4) array.
        '''
        color_keys = ['red', 'green', 'blue', 'alpha']
        candidate_colors = [element['data'][i] for i in color_keys if i in element['properties']]

        if len(candidate_colors) >= 3:
            return np.column_stack(candidate_colors)
        return None

    def ply_read_header(file_obj):
        '''
        Read the ASCII header of a PLY file, and leave the file object 
        at the position of the start of data. 
        '''
        # from ply specification
        dtypes = {'char'  : 'i1',
                  'uchar' : 'u1',
                  'short' : 'i2',
                  'ushort': 'u2',
                  'int'   : 'i4',
                  'uint'  : 'u4',
                  'float' : 'f4',
                  'double': 'f8'}

        if not 'ply' in str(file_obj.readline()):
            raise ValueError('This aint a ply file')
        encoding = str(file_obj.readline()).strip().split()[1]

        if 'ascii' in encoding:
            raise ValueError('ASCII PLY not supported!')

        endian   = ['<', '>']['big' in encoding]
        elements = OrderedDict()

        while True:
            line = file_obj.readline()
            if line is None:
                raise ValueError('Header wasn\'t terminated properly!')
            line = line.decode('utf-8').strip().split()
            
            if 'end_header' in line:
                break

            if 'element' in line[0]:
                name, length = line[1:]
                elements[name] = {'length'     : int(length), 
                                  'properties' : OrderedDict()}
            elif 'property' in line[0]:
                if len(line) == 3:
                    dtype, field = line[1:]
                    elements[name]['properties'][str(field)] = endian + dtypes[dtype]
                elif 'list' in line[1]:
                    dtype_count, dtype, field = line[2:]
                    elements[name]['properties'][str(field)] = (endian +
                                                                dtypes[dtype_count] + 
                                                                ', ($LIST,)'+
                                                                endian + 
                                                                dtypes[dtype])
        return elements

    def ply_populate_listsize(file_obj, elements):
        '''
        Given a set of elements populated from the header if there are any
        list properties seek in the file the length of the list. 

        Note that if you have a list where each instance is different length
        (if for example you mixed triangles and quads) this won't work at all
        '''
        p_start = file_obj.tell()
        p_current = file_obj.tell()
        for element_key, element in elements.items():
            props = element['properties']
            prior_data = ''
            for k, dtype in props.items():
                if '$LIST' in dtype:
                    # every list field has two data types: 
                    # the list length (single value), and the list data (multiple)
                    # here we are only reading the single value for list length
                    field_dtype = np.dtype(dtype.split(',')[0])
                    if len(prior_data) == 0: 
                        offset = 0
                    else: 
                        offset = np.dtype(prior_data).itemsize()
                    file_obj.seek(p_current+offset)
                    size = np.fromstring(file_obj.read(field_dtype.itemsize), 
                                         dtype=field_dtype)[0]
                    props[k] = props[k].replace('$LIST', str(size))
                prior_data += props[k] +','
            itemsize = np.dtype(', '.join(props.values())).itemsize
            p_current += element['length'] * itemsize
        file_obj.seek(p_start)
        
    def ply_populate_data(file_obj, elements):
        '''
        Given the data type and field information from the header,
        read the data and add it to a 'data' field in the element.
        '''
        for key in elements.keys():
            items = list(elements[key]['properties'].items())
            dtype = np.dtype(items)
            data  = file_obj.read(elements[key]['length'] * dtype.itemsize)
            elements[key]['data'] = np.fromstring(data, dtype=dtype)
        return elements

    def ply_elements_kwargs(elements):
        '''
        Given an elements data structure, extract the keyword
        arguments that a Trimesh object constructor will expect.
        '''
        vertices = np.column_stack([elements['vertex']['data'][i] for i in 'xyz'])
        faces    = elements['face']['data']['vertex_indices']['f1']
        face_colors   = ply_element_colors(elements['face'])
        vertex_colors = ply_element_colors(elements['vertex'])
        result = {'vertices' : vertices,
                  'faces'    : faces,
                  'face_colors'   : face_colors,
                  'vertex_colors' : vertex_colors}
        return result

    def ply_elements_size(elements):
        '''
        Given an elements data structure populated from the header, 
        calculate how long the file should be if it is intact.
        '''
        size = 0
        for element in elements.values():
            dtype = np.dtype(','.join(element['properties'].values()))
            size += element['length'] * dtype.itemsize
        return size
    
    # OrderedDict which is populated from the header
    elements = ply_read_header(file_obj)
    # some elements are passed where the list dimensions
    # are not included in the header, so this function goes 
    # into the meat of the file and grabs the list dimensions 
    # before we to the main data read as a single operation
    ply_populate_listsize(file_obj, elements)

    # how many bytes are left in the file
    size_file = size_to_end(file_obj)
    # how many bytes should the data structure described by
    # the header take up
    size_elements = ply_elements_size(elements)
    
    # if the number of bytes is not the same the file is probably corrupt
    if size_file != size_elements:
        raise ValueError('File is unexpected length!')

    # with everything populated and a reasonable confidence the file
    # is intact, read the data fields described by the header
    ply_populate_data(file_obj, elements)
    # all of the data is now stored in elements, but we need it as
    # a set of keyword arguments we can pass to the Trimesh constructor
    # will look something like {'vertices' : (data), 'faces' : (data)} 
    mesh_kwargs = ply_elements_kwargs(elements)
    return mesh_kwargs

def export_ply(mesh):
    '''
    Export a mesh in the PLY format.
    
    Arguments
    ----------
    mesh: Trimesh object
    
    Returns
    ----------
    export: bytes of result
    '''
    dtype_face = np.dtype([('count', '<u1'), 
                           ('index', '<i4', (3))])
    dtype_vertex = np.dtype([('vertex', '<f4', (3))])
    
    faces = np.zeros(len(mesh.faces), dtype=dtype_face)
    faces['count'] = 3
    faces['index'] = mesh.faces
    
    vertex = np.zeros(len(mesh.vertices), dtype=dtype_vertex)
    vertex['vertex'] = mesh.vertices
    
    template = Template(get_resource('ply.template'))
    export = template.substitute({'vertex_count' : len(mesh.vertices),
                                  'face_count'   : len(mesh.faces)}).encode('utf-8')
    export += vertex.tostring()
    export += faces.tostring()
    return export
    


def size_to_end(file_obj):
    '''
    Given an open file object, return the number of bytes 
    to the end of the file
    '''
    position_current = file_obj.tell()
    file_obj.seek(0,2)
    position_end = file_obj.tell()
    file_obj.seek(position_current)
    size = position_end - position_current
    return size

_ply_loaders = {'ply' : load_ply}



import numpy as np

from . import transformations
from . import util
from . import points

def transform(mesh):
    '''
    Return a permutated variant of a mesh by randomly reording faces 
    and rotatating + translating a mesh by a random matrix.
    
    Arguments
    ----------
    mesh:   Trimesh object (will not be altered by this function)
             
    Returns
    ----------
    permutated: Trimesh object, same faces as input mesh but
                rotated and reordered.
    '''
    matrix = transformations.random_rotation_matrix()
    matrix[0:3,3] = np.random.random(3)*1000

    triangles = np.random.permutation(mesh.triangles).reshape((-1,3))
    triangles = points.transform_points(triangles, matrix)

    mesh_type = util.type_named(mesh, 'Trimesh')
    permutated = mesh_type(vertices = triangles,
                           faces    = np.arange(len(mesh.faces)*3).reshape((-1,3)))

    return permutated
    
def noise(mesh, magnitude=None):
    '''
    Add gaussian noise to every vertex of a mesh.
    Makes no effort to maintain topology or sanity.
    
    Arguments
    ----------
    mesh:      Trimesh object (will not be mutated)
    magnitude: float, what is the maximum distance per axis we can displace a vertex. 
               Default value is mesh.scale/100.0
           
    Returns
    ----------
    permutated: Trimesh object, input mesh with noise applied
    '''
    if magnitude is None:
        magnitude = mesh.scale / 100.0
    
    # make sure we've re- ordered faces randomly
    faces = np.random.permutation(mesh.faces)
    vertices = mesh.vertices + ((np.random.random(mesh.vertices.shape) - .5) * magnitude)
    
    mesh_type = util.type_named(mesh, 'Trimesh')
    permutated = mesh_type(vertices = vertices,
                           faces    = faces)
    return permutated
    
def tesselation(mesh):
    '''
    Subdivide each face of a mesh into three faces with the new vertex
    randomly placed inside the old face. 
    
    This produces a mesh with exactly the same surface area and volume 
    but with different tesselation.  
    
    Arguments
    ----------
    mesh: Trimesh object
    
    Returns
    ----------
    permutated: Trimesh object with remeshed facets
    '''
    # create random barycentric coordinates for each face
    # pad all coordinates by a small amount to bias new vertex towards center
    barycentric  = np.random.random(mesh.faces.shape) + .05
    barycentric /=  barycentric.sum(axis=1).reshape((-1,1))

    # create one new vertex somewhere in a face
    vertex_face = (barycentric.reshape((-1,3,1)) * mesh.triangles).sum(axis=1)
    vertex_face_id = np.arange(len(vertex_face)) + len(mesh.vertices)
    
    # new vertices are the old vertices stacked on the vertices in the faces
    vertices = np.vstack((mesh.vertices, vertex_face))
    # there are three new faces per old face, and we maintain correct winding
    faces = np.vstack((np.column_stack((mesh.faces[:,[0,1]], vertex_face_id)),
                       np.column_stack((mesh.faces[:,[1,2]], vertex_face_id)),
                       np.column_stack((mesh.faces[:,[2,0]], vertex_face_id))))
    # make sure the order of the faces is permutated
    faces = np.random.permutation(faces)
    
    mesh_type = util.type_named(mesh, 'Trimesh')
    permutated = mesh_type(vertices = vertices,
                           faces    = faces)
    return permutated
    
class Permutator:
    def __init__(self, mesh):
        '''
        A convienence object to get permutated versions of a mesh.
        '''
        self._mesh = mesh
    def transform(self):
        return transform(self._mesh)
    def noise(self, magnitude=None):
        return noise(self._mesh, magnitude)
    def tesselation(self):
        return tesselation(self._mesh)


import numpy as np

from . import util
from . import points
from . import creation

from .base      import Trimesh
from .constants import log
from .triangles import windings_aligned
   
class _Primitive(Trimesh):
    '''
    Geometric _Primitives which are a subclass of Trimesh.
    Mesh is generated lazily when vertices or faces are requested.
    '''
    def __init__(self, *args, **kwargs):
        super(_Primitive, self).__init__(*args, **kwargs)
        self._data.clear()
        self._validate = False
        
    @property
    def faces(self):
        stored = self._cache['faces']
        if util.is_shape(stored, (-1,3)):
            return stored
        self._create_mesh()
        #self._validate_face_normals()
        return self._cache['faces']

    @faces.setter
    def faces(self, values):
        log.warning('Primitive faces are immutable! Not setting!')

    @property
    def vertices(self):
        stored = self._cache['vertices']
        if util.is_shape(stored, (-1,3)):
            return stored

        self._create_mesh()
        return self._cache['vertices']

    @vertices.setter
    def vertices(self, values):
        if values is not None:
            log.warning('Primitive vertices are immutable! Not setting!')

    @property
    def face_normals(self):
        stored = self._cache['face_normals']
        if util.is_shape(stored, (-1,3)):
            return stored
        self._create_mesh()
        return self._cache['face_normals']
    
    @face_normals.setter
    def face_normals(self, values):
        if values is not None:
            log.warning('Primitive face normals are immutable! Not setting!')

    def to_mesh(self):
        '''
        Return a copy of the _Primitive object as a trimesh.
        '''
        result = Trimesh(vertices = self.vertices,
                         faces    = self.faces,
                         face_normals = self.face_normals)
        return result
            
    def _create_mesh(self):
        raise ValueError('Primitive doesn\'t define mesh creation!')

class _PrimitiveAttributes(object):
    '''
    Store Primitive attributes. 
    '''
    def __init__(self, data, defaults, kwargs):
        self._data     = data
        self._defaults = defaults
        self._data.update(defaults)
        
        for key, value in kwargs.items():
            if key in defaults:
                self._data[key] = util.convert_like(value, defaults[key])
        
    def __getattr__(self, key):
        if '_' in key:
            return super(_PrimitiveAttributes, self).__getattr__(key)
        elif key in self._defaults:
            return util.convert_like(self._data[key], self._defaults[key])
        return super(_PrimitiveAttributes, self).__getattr__(key)
            
    def __setattr__(self, key, value):
        if '_' in key:
            return super(_PrimitiveAttributes, self).__setattr__(key, value) 
        elif key in self._defaults:
            self._data[key] = value
        else:
            raise ValueError('Only default attributes {} can be set!'.format(list(self._defaults.keys())))
            
class Cylinder(_Primitive):
    def __init__(self, *args, **kwargs):
        '''
        Create a Sphere _Primitive, which is a subclass of Trimesh.

        Arguments
        ----------
        sphere_radius: float, radius of sphere
        sphere_center: (3,) float, center of sphere
        subdivisions: int, number of subdivisions for icosphere. Default is 3
        '''
        super(Cylinder, self).__init__(*args, **kwargs)
        
        defaults = {'height'    : 1.0,
                    'radius'    : 1.0,
                    'transform' : np.eye(4),
                    'sections'  : 32}

        self.primitive_data = _PrimitiveAttributes(self._data, defaults, kwargs)

    def _create_mesh(self):
        log.info('Creating cylinder mesh with r=%f, h=%f and %d sections',
                 self.primitive_data.radius,
                 self.primitive_data.height,
                 self.primitive_data.sections)
                 
        mesh = creation.cylinder(radius   = self.primitive_data.radius,
                                 height   = self.primitive_data.height,
                                 sections = self.primitive_data.sections)
        mesh.apply_transform(self.primitive_data.transform)
        
        self._cache['vertices']     = mesh.vertices
        self._cache['faces']        = mesh.faces
        self._cache['face_normals'] = mesh.face_normals
        
        
class Sphere(_Primitive):
    def __init__(self, *args, **kwargs):
        '''
        Create a Sphere _Primitive, which is a subclass of Trimesh.

        Arguments
        ----------
        sphere_radius: float, radius of sphere
        sphere_center: (3,) float, center of sphere
        subdivisions: int, number of subdivisions for icosphere. Default is 3
        '''
        try:    subdivisions = int(kwargs['subdivisions'])
        except: subdivisions = 3

        self._unit_sphere = creation.icosphere(subdivisions=subdivisions)
        super(Sphere, self).__init__(*args, **kwargs)

        if 'sphere_radius' in kwargs:
            self.sphere_radius = kwargs['sphere_radius']
        if 'sphere_center' in kwargs:
            self.sphere_center = kwargs['sphere_center']

    @property
    def sphere_center(self):
        stored = self._data['center']
        if stored is None:
            return np.zeros(3)
        return stored

    @sphere_center.setter
    def sphere_center(self, values):
        self._data['center'] = np.asanyarray(values, dtype=np.float64)

    @property
    def sphere_radius(self):
        stored = self._data['radius']
        if stored is None:
            return 1.0
        return stored[0]

    @sphere_radius.setter
    def sphere_radius(self, value):
        self._data['radius'] = float(value)

    @util.cache_decorator
    def volume(self):
        '''
        Volume of the current sphere _Primitive.

        Returns
        --------
        volume: float, volume of the sphere _Primitive
        '''
        
        volume = (4.0*np.pi * (self.sphere_radius ** 3)) / 3.0
        return volume
    
    def _create_mesh(self):
        ico = self._unit_sphere
        self._cache['vertices'] = ((ico.vertices * self.sphere_radius) + 
                                   self.sphere_center)
        self._cache['faces'] = ico.faces
        self._cache['face_normals'] = ico.face_normals

class Box(_Primitive):    
    def __init__(self, *args, **kwargs):
        '''
        Create a Box _Primitive, which is a subclass of Trimesh

        Arguments
        ----------
        box_extents:   (3,) float, size of box
        box_transform: (4,4) float, transformation matrix for box
        box_center:    (3,) float, convience function which updates box_transform
                       with a translation- only matrix
        '''
        self._unit_box = creation.box()
        super(Box, self).__init__(*args, **kwargs)

        if 'box_extents' in kwargs:
            self.box_extents = kwargs['box_extents']
        if 'box_transform' in kwargs:
            self.box_transform = kwargs['box_transform']
        if 'box_center' in kwargs:
            self.box_center = kwargs['box_center']

    @property
    def box_center(self):
        return self.box_transform[0:3,3]

    @box_center.setter
    def box_center(self, values):
        transform = self.box_transform
        transform[0:3,3] = values
        self._data['box_transform'] = transform

    @property
    def box_extents(self):
        stored = self._data['box_extents']
        if util.is_shape(stored, (3,)):
            return stored
        return np.ones(3)

    @box_extents.setter
    def box_extents(self, values):
        self._data['box_extents'] = np.asanyarray(values, dtype=np.float64)

    @property
    def box_transform(self):
        stored = self._data['box_transform']
        if util.is_shape(stored, (4,4)):
            return stored
        return np.eye(4)

    @box_transform.setter
    def box_transform(self, matrix):
        matrix = np.asanyarray(matrix, dtype=np.float64)
        if matrix.shape != (4,4):
            raise ValueError('Matrix must be (4,4)!')
        self._data['box_transform'] = matrix

    @property
    def is_oriented(self):
        if util.is_shape(self.box_transform, (4,4)):
            return not np.allclose(self.box_transform[0:3,0:3], np.eye(3))
        else:
            return False

    @util.cache_decorator
    def volume(self):
        '''
        Volume of the box _Primitive.

        Returns
        --------
        volume: float, volume of box
        '''
        volume = float(np.product(self.box_extents))
        return volume
        
    def _create_mesh(self):
        log.debug('Creating mesh for box _Primitive')
        box = self._unit_box
        vertices, faces, normals = box.vertices, box.faces, box.face_normals
        vertices = points.transform_points(vertices * self.box_extents, 
                                           self.box_transform)
        normals = np.dot(self.box_transform[0:3,0:3], 
                         normals.T).T
        aligned = windings_aligned(vertices[faces[:1]], normals[:1])[0]
        if not aligned:
            faces = np.fliplr(faces)        
        # for a _Primitive the vertices and faces are derived from other information
        # so it goes in the cache, instead of the datastore
        self._cache['vertices'] = vertices
        self._cache['faces']    = faces
        self._cache['face_normals'] = normals

class Extrusion(_Primitive):
    def __init__(self, *args, **kwargs):
        '''
        Create an Extrusion _Primitive, which subclasses Trimesh

        Arguments
        ----------
        extrude_polygon:   shapely.geometry.Polygon, polygon to extrude
        extrude_transform: (4,4) float, transform to apply after extrusion
        extrude_height:    float, height to extrude polygon by
        '''
        super(Extrusion, self).__init__(*args, **kwargs)
        if 'extrude_polygon' in kwargs:
            self.extrude_polygon   = kwargs['extrude_polygon']
        if 'extrude_transform' in kwargs:
            self.extrude_transform = kwargs['extrude_transform']
        if 'extrude_height' in kwargs:
            self.extrude_height    = kwargs['extrude_height']

    @property
    def extrude_transform(self):
        stored = self._data['extrude_transform']
        if np.shape(stored) == (4,4):
            return stored
        return np.eye(4)

    @extrude_transform.setter
    def extrude_transform(self, matrix):
        matrix = np.asanyarray(matrix, dtype=np.float64)
        if matrix.shape != (4,4):
            raise ValueError('Matrix must be (4,4)!')
        self._data['extrude_transform'] = matrix

    @property
    def extrude_height(self):
        stored = self._data['extrude_height']
        if stored is None: 
            raise ValueError('extrude height not specified!')
        return stored.copy()[0]

    @extrude_height.setter
    def extrude_height(self, value):
        self._data['extrude_height'] = float(value)

    @property
    def extrude_polygon(self):
        stored = self._data['extrude_polygon']
        if stored is None: 
            raise ValueError('extrude polygon not specified!')
        return stored[0]

    @extrude_polygon.setter
    def extrude_polygon(self, value):
        polygon = creation.validate_polygon(value)
        self._data['extrude_polygon'] = polygon
        
    @property
    def extrude_direction(self):
        direction = np.dot(self.extrude_transform[:3,:3], 
                           [0.0,0.0,1.0])
        return direction
    
    def slide(self, distance):
        distance = float(distance)
        translation = np.eye(4)
        translation[2,3] = distance
        new_transform = np.dot(self.extrude_transform.copy(),
                               translation.copy())
        self.extrude_transform = new_transform

    def _create_mesh(self):
        log.debug('Creating mesh for extrude _Primitive')
        mesh = creation.extrude_polygon(self.extrude_polygon,
                                        self.extrude_height)
        mesh.apply_transform(self.extrude_transform)
        self._cache['vertices']     = mesh.vertices
        self._cache['faces']        = mesh.faces
        self._cache['face_normals'] = mesh.face_normals
